1#ifndef AMREX_FFT_OPENBC_SOLVER_H_
2#define AMREX_FFT_OPENBC_SOLVER_H_
24template <
typename T = Real>
61 [[nodiscard]]
Box const&
Domain ()
const {
return m_domain; }
94 static IntVect make_padded_length (
Box const& domain,
Info const& info);
95 static Box make_grown_domain (
Box const& domain,
IntVect const& padded_len,
103 std::unique_ptr<R2C<T>> m_r2c_green;
110 int ndims = AMREX_SPACEDIM;
111#if (AMREX_SPACEDIM == 3)
112 if (info.twod_mode) { ndims = 2; }
116 if (info.openbc_padding) {
117 for (
int idim = 0; idim < ndims; ++idim) {
125Box OpenBCSolver<T>::make_grown_domain (
Box const& domain,
IntVect const& padded_len,
129 int ndims = AMREX_SPACEDIM;
130#if (AMREX_SPACEDIM == 3)
131 if (info.twod_mode) { ndims = 2; }
135 for (
int idim = 0; idim < ndims; ++idim) {
137 len[idim] <= std::numeric_limits<int>::max()/2,
138 "FFT::OpenBCSolver: padded domain length exceeds int range");
141 return Box(domain.smallEnd(), domain.smallEnd()+len-
IntVect(1), domain.ixType());
148 m_padded_length(
OpenBCSolver<T>::make_padded_length(domain, info)),
149 m_r2c(
OpenBCSolver<T>::make_grown_domain(domain, m_padded_length, info),
153 "FFT::OpenBCSolver does not support FFT::Info::batch_size > 1");
155#if (AMREX_SPACEDIM == 3)
157 auto gdom = make_grown_domain(domain, m_padded_length, m_info);
158 gdom.enclosedCells(2);
163 gdom.setBig(2, nprocs-1);
164 m_r2c_green = std::make_unique<R2C<T>>(gdom,m_info);
165 auto [sd, ord] = m_r2c_green->getSpectralData();
171 auto [sd, ord] = m_r2c.getSpectralData();
173 m_G_fft.define(sd->boxArray(), sd->DistributionMap(), 1, 0);
181 BL_PROFILE(
"OpenBCSolver::setGreensFunction");
183 auto* infab = m_info.twod_mode ? detail::get_fab(m_r2c_green->m_rx)
184 : detail::get_fab(m_r2c.m_rx);
185 auto const& lo = m_domain.smallEnd();
186 auto const& lo3 = lo.dim3();
187 auto const len3d = m_padded_length.dim3();
190 auto const& a = infab->array();
191 auto box = infab->box();
193 int ndims = m_info.twod_mode ? AMREX_SPACEDIM-1 : AMREX_SPACEDIM;
194 for (
int idim = 0; idim < ndims; ++idim) {
195 if (box.smallEnd(idim) == lo[idim] && box.length(idim) == 2*len[idim]) {
196 box.growHi(idim, -len[idim]+1);
205 if (i == len[0] || j == len[1] || k == len[2]) {
209 auto jj = (j > len[1]) ? 2*len[1]-j : j;
210 auto kk = (k > len[2]) ? 2*len[2]-k : k;
211 G = greens_function(ii+lo3.x,jj+lo3.y,kk+lo3.z);
213 for (
int koff = 0; koff < nimages[2]; ++koff) {
214 int k2 = (koff == 0) ? k : 2*len[2]-k;
215 if ((k2 == 2*len[2]) || (koff == 1 && k == len[2])) {
218 for (
int joff = 0; joff < nimages[1]; ++joff) {
219 int j2 = (joff == 0) ? j : 2*len[1]-j;
220 if ((j2 == 2*len[1]) || (joff == 1 && j == len[1])) {
223 for (
int ioff = 0; ioff < nimages[0]; ++ioff) {
224 int i2 = (ioff == 0) ? i : 2*len[0]-i;
225 if ((i2 == 2*len[0]) || (ioff == 1 && i == len[0])) {
228 a(i2+lo3.x,j2+lo3.y,k2+lo3.z) = G;
235 if (m_info.twod_mode) {
236 m_r2c_green->forward(m_r2c_green->m_rx);
238 m_r2c.forward(m_r2c.m_rx);
241 if (!m_info.twod_mode) {
242 auto [sd, ord] = m_r2c.getSpectralData();
244 auto const* srcfab = detail::get_fab(*sd);
246 auto* dstfab = detail::get_fab(m_G_fft);
254 m_r2c.prepare_openbc();
263 auto& inmf = m_r2c.m_rx;
265 inmf.ParallelCopy(rho, 0, 0, 1);
267 m_r2c.m_openbc_half = !m_info.twod_mode;
269 m_r2c.m_openbc_half =
false;
271 auto scaling_factor = m_r2c.scalingFactor();
273 auto const* gfab = detail::get_fab(m_G_fft);
275 auto [sd, ord] = m_r2c.getSpectralData();
277 auto* rhofab = detail::get_fab(*sd);
279 auto*
pdst = rhofab->dataPtr();
280 auto const* psrc = gfab->dataPtr();
281 Box const& rhobox = rhofab->box();
282#if (AMREX_SPACEDIM == 3)
284 if (m_info.twod_mode) {
293#if (AMREX_SPACEDIM == 3)
294 Long isrc = i % leng;
298 pdst[i] *= psrc[isrc] * scaling_factor;
301 amrex::Abort(
"FFT::OpenBCSolver::solve: how did this happen?");
305 m_r2c.m_openbc_half = !m_info.twod_mode;
306 m_r2c.backward_doit(phi, phi.nGrowVect());
307 m_r2c.m_openbc_half =
false;
#define BL_PROFILE(a)
Definition AMReX_BLProfiler.H:551
#define AMREX_ALWAYS_ASSERT_WITH_MESSAGE(EX, MSG)
Definition AMReX_BLassert.H:49
#define AMREX_ASSERT(EX)
Definition AMReX_BLassert.H:38
#define AMREX_GPU_DEVICE
Definition AMReX_GpuQualifiers.H:18
Real * pdst
Definition AMReX_HypreMLABecLap.cpp:1130
__host__ __device__ Long numPts() const noexcept
Return the number of points contained in the BoxND.
Definition AMReX_Box.H:385
__host__ __device__ IntVectND< dim > length() const noexcept
Return the length of the BoxND.
Definition AMReX_Box.H:167
Convolution-based solver for open boundary conditions using Green's functions.
Definition AMReX_FFT_OpenBCSolver.H:26
cMF const & greensFunctionFFT() const noexcept
Access the spectral Green's function held by this solver.
Definition AMReX_FFT_OpenBCSolver.H:91
Box const & Domain() const
Access the physical domain this solver was built for.
Definition AMReX_FFT_OpenBCSolver.H:61
typename R2C< T >::MF MF
Definition AMReX_FFT_OpenBCSolver.H:28
void solve(MF &phi, MF const &rho)
Solve for phi given right-hand side rho.
Definition AMReX_FFT_OpenBCSolver.H:259
void setGreensFunction(F const &greens_function)
Populate the spectral Green's function used by subsequent solves.
Definition AMReX_FFT_OpenBCSolver.H:179
IntVect const & PaddedLength() const
Access the one-sided padded length used to build the internal FFT domain.
Definition AMReX_FFT_OpenBCSolver.H:68
typename R2C< T >::cMF cMF
Definition AMReX_FFT_OpenBCSolver.H:29
OpenBCSolver(Box const &domain, Info const &info=Info{})
Build a solver over domain using the FFT Info settings in info.
Definition AMReX_FFT_OpenBCSolver.H:145
cMF & greensFunctionFFT() noexcept
Access the spectral Green's function held by this solver.
Definition AMReX_FFT_OpenBCSolver.H:84
Parallel Discrete Fourier Transform.
Definition AMReX_FFT_R2C.H:48
std::conditional_t< C, cMF, std::conditional_t< std::is_same_v< T, Real >, MultiFab, FabArray< BaseFab< T > > > > MF
Definition AMReX_FFT_R2C.H:53
Open Boundary Poisson Solver.
Definition AMReX_OpenBC.H:70
OpenBCSolver()=default
Construct an empty solver; call define() before solving.
amrex_long Long
Definition AMReX_INT.H:30
void ParallelForOMP(T n, L const &f) noexcept
Performance-portable kernel launch function with optional OpenMP threading.
Definition AMReX_GpuLaunch.H:328
Definition AMReX_FFT_Helper.H:53
int nextFastLen(int target, int nfactors=FastNumPrimeFactors())
Return the smallest fast FFT length greater than or equal to target.
Definition AMReX_FFT_Helper.H:286
DomainStrategy
Definition AMReX_FFT_Helper.H:57
void dtod_memcpy_async(void *p_d_dst, const void *p_d_src, const std::size_t sz) noexcept
Definition AMReX_GpuDevice.H:449
int NProcsSub() noexcept
number of ranks in current frame
Definition AMReX_ParallelContext.H:74
@ make_alias
Definition AMReX_MakeType.H:7
__host__ __device__ void ignore_unused(const Ts &...)
No-op helper that marks variables as intentionally unused.
Definition AMReX.H:259
BoxND< 3 > Box
Box is an alias for amrex::BoxND instantiated with AMREX_SPACEDIM.
Definition AMReX_BaseFwd.H:35
IntVectND< 3 > IntVect
IntVect is an alias for amrex::IntVectND instantiated with AMREX_SPACEDIM.
Definition AMReX_BaseFwd.H:38
void Abort(const std::string &msg)
Print a fatal-error message to stderr and abort execution.
Definition AMReX.cpp:242
Definition AMReX_FFT_Helper.H:83
bool twod_mode
Definition AMReX_FFT_Helper.H:94
int batch_size
Batched FFT size. Only support in R2C, not R2X.
Definition AMReX_FFT_Helper.H:106
int nprocs
Max number of processes to use.
Definition AMReX_FFT_Helper.H:109
Fixed-size array that can be used on GPU.
Definition AMReX_Array.H:52